"""Unit tests for beanbag_docutils.sphinx.ext.image_srcsets."""

from beanbag_docutils.sphinx.ext import image_srcsets
from beanbag_docutils.sphinx.ext.tests.testcase import SphinxExtTestCase


class ImageTests(SphinxExtTestCase):
    """Unit tests for image."""

    extensions = [
        image_srcsets.__name__,
    ]

    def test_with_html(self):
        """Testing image with HTML"""
        rendered = self.render_doc(
            '.. image:: path/to/image.png\n'
        )

        self.assertEqual(
            rendered,
            '<img alt="path/to/image.png" src="path/to/image.png" />')

    def test_with_html_and_srcset(self):
        """Testing image-srcset with HTML and srcset"""
        rendered = self.render_doc(
            '.. image:: path/to/image.png\n'
            '   :sources: 2x path/to/image@2x.png\n'
            '             3x path/to/image@3x.png\n'
        )

        self.assertEqual(
            rendered,
            '<img srcset="_images/image.png 1x, _images/image%402x.png 2x,'
            ' _images/image%403x.png 3x" alt="_images/image.png"'
            ' src="_images/image.png" />')

    def test_with_html_and_srcset_files(self):
        """Testing image-srcset with HTML and srcset @-based files"""
        rendered = self.render_doc(
            '.. image:: my/images/image.png\n',
            extra_files={
                'my/images/image.png': b'',
                'my/images/image@2x.png': b'',
                'my/images/image@3x.png': b'',
                'my/images/image@4x.png': b'',
            })

        self.assertEqual(
            rendered,
            '<img srcset="_images/image.png 1x, _images/image%402x.png 2x,'
            ' _images/image%403x.png 3x, _images/image%404x.png 4x"'
            ' alt="_images/image.png" src="_images/image.png" />')

    def test_with_html_and_width_height(self):
        """Testing image-srcset with HTML and srcset and width/height"""
        rendered = self.render_doc(
            '.. image:: path/to/image.png\n'
            '   :width: 100\n'
            '   :height: 200\n'
            '   :sources: 2x path/to/image@2x.png\n'
            '             3x path/to/image@3x.png\n'
        )

        self.assertEqual(
            rendered,
            '<a class="reference internal image-reference"'
            ' href="_images/image.png">'
            '<img srcset="_images/image.png 1x, _images/image%402x.png 2x,'
            ' _images/image%403x.png 3x" width="100" height="200"'
            ' alt="_images/image.png" src="_images/image.png" />'
            '</a>')



"""Base test case support for Sphinx extensions."""

import os
from contextlib import contextmanager
from unittest import TestCase
from typing import Dict, Iterator

import six
from sphinx_testing.util import TestApp, docutils_namespace

import beanbag_docutils


_theme_path = os.path.abspath(os.path.join(beanbag_docutils.__file__,
                                           '..', '..', 'tests', 'sphinx',
                                           'themes'))


class SphinxExtTestCase(TestCase):
    """Base class for test cases for Sphinx extensions.

    This provides common functionality for setting up and rendering a ReST
    document to a content-only HTML string.
    """

    #: Class-wide configuration for the renders.
    config = {}

    #: A list of extensions required for the tests.
    extensions = []

    maxDiff = 1000000

    @contextmanager
    def with_sphinx_env(self, config={}, builder_name='html'):
        """Run within a Sphinx environment.

        Version Added:
            1.10

        Args:
            config (dict, optional):
                Additional configuration to set for phinx.

        Context:
            dict:
            A dictionary containing:

            Keys:
                app (sphinx.application.Sphinx):
                    The Sphinx application that was set up.

                config (sphinx.config.Config):
                    The Sphinx configuration object.

                srcdir (unicode):
                    The location to place ReST source files in.
        """
        new_config = {
            'extensions': self.extensions,
            'html_theme': 'test',
            'html_theme_path': [_theme_path],
        }
        new_config.update(self.config)
        new_config.update(config)

        with docutils_namespace():
            app = TestApp(buildername=builder_name,
                          create_new_srcdir=True,
                          copy_srcdir_to_tmpdir=False,
                          confoverrides=new_config)

            try:
                yield {
                    'app': app,
                    'config': app.config,
                    'srcdir': app.srcdir,
                }
            finally:
                app.cleanup()

    @contextmanager
    def rendered_docs(
        self,
        files={},            # type: Dict
        config={},           # type: Dict
        builder_name='html'  # type: str
    ):  # type: (...) -> Iterator[str]
        """Render ReST documents and yield the directory for processing.

        This will set up a Sphinx environment, based on the extensions and
        configuration provided by the consumer, and perform a render of the
        given ReST files. The caller can then inspect the files within the
        environment.

        Version Added:
            2.2

        Args:
            files (dict):
                A mapping of filenames and contents to write.

                All contents are byte strings.

            config (dict, optional):
                Additional configuration to set for the render.

            builder_name (unicode, optional):
                The name of the builder to use.

        Context:
            unicode:
        """
        with self.with_sphinx_env(config=config,
                                  builder_name=builder_name) as ctx:
            srcdir = ctx['srcdir']

            old_cwd = os.getcwd()
            os.chdir(srcdir)

            try:
                for path, contents in files.items():
                    dirname = os.path.dirname(path)

                    if dirname and not os.path.exists(dirname):
                        os.makedirs(dirname)

                    if isinstance(contents, bytes):
                        with open(path, 'wb') as fp:
                            fp.write(contents)
                    else:
                        assert isinstance(contents, six.text_type)

                        with open(path, 'w') as fp:
                            fp.write(contents)

                app = ctx['app']
                app.build()

                yield str(app.outdir)
            finally:
                os.chdir(old_cwd)

    def render_doc(
        self,
        doc_content: str,
        config={},
        builder_name='html',
        extra_files={}
    ):  # type: (...) -> str
        """Render a ReST document to a string.

        This will set up a Sphinx environment, based on the extensions and
        configuration provided by the consumer, and perform a render of the
        given ReST content. The resulting string will be a rendered version
        of just that content.

        Args:
            doc_content (unicode):
                The ReST content to render.

            config (dict, optional):
                Additional configuration to set for the render.

            builder_name (unicode, optional):
                The name of the builder to use.

            extra_files (dict):
                A mapping of extra filenames and contents to write.

        Returns:
            unicode:
            The rendered content.

        Raises:
            ValueError:
                ``builder_name`` wasn't a valid value.
        """
        if builder_name == 'html':
            out_filename = 'contents.html'
        elif builder_name == 'json':
            out_filename = 'contents.fjson'
        else:
            raise ValueError('"%s" is not a supported builder name'
                             % builder_name)

        files = {
            'contents.rst': doc_content,
        }
        files.update(extra_files)

        with self.rendered_docs(files=files,
                                config=config,
                                builder_name=builder_name) as build_dir:
            with open(os.path.join(build_dir, out_filename), 'r') as fp:
                return fp.read().strip()



"""Sphinx extension for srcsets in images.

.. versionadded:: 2.1


This extension adds a ``sources`` option to the standard image directives,
enabling responsive image support via srcsets.

These are specified a bit differently from ``<img srcset="...">`` values.
The descriptor goes first, and a comma between entries is optional (a blank
line can be used instead). For example:

.. code-block:: rst

   .. image:: path/to/file.png
      :sources: 2x path/to/file@2x.png
                3x path/to/file@3x.png
                100w path/to/file@100w.png
                200h path/to/file@200h.png

If ``sources`` is not explicitly provided, but files with those standard
``@``-based suffixes exist alongside the referenced main image, they'll
automatically be used to define the srcsets of the image. The ``1x`` entry is
also automatically inserted based on the main image.

If relying on the default of scanning for srcset images, this becomes a
zero-configuration, drop-in solution for all Sphinx documentation codebases.


Setup
=====

To use this, you just need to add the extension in :file:`conf.py`::

    extensions = [
        ...
        'beanbag_docutils.sphinx.ext.image_srcsets',
        ...
    ]
"""

import os
import posixpath
import re
from collections import OrderedDict
from glob import glob
from typing import Dict, List, Optional, TYPE_CHECKING

from docutils import nodes
from docutils.parsers.rst import directives
from docutils.parsers.rst.directives.images import Image
from six.moves.urllib.parse import quote as urllib_quote
from sphinx.application import Sphinx
from sphinx.util.i18n import search_image_for_language

from beanbag_docutils import VERSION

if TYPE_CHECKING:
    from sphinx.environment import BuildEnvironment


WIDTH_ATTR_RE = re.compile(r' width="(\d+)"')
HEIGHT_ATTR_RE = re.compile(r' height="(\d+)"')
STYLE_ATTR_RE = re.compile(r' style="(?P<style>[^"]+)"')


def _get_srcsets(
    env,          # type: BuildEnvironment
    node,         # type: nodes.image
    docname=None  # type: Optional[str]
):  # type: (...) -> Dict[str, str]
    """Return a normalized version of all srcsets for an image node.

    This will convert a ``sources`` option to a dictionary mapping descriptors
    (such as ``2x``, ``100w``, etc.) to URLs.

    These will be cached for future lookup.

    Args:
        env (sphinx.environment.BuildEnvironment):
            The current Sphinx build environment.

        node (docutils.nodes.image):
            The image node to retrieve sources from.

        docname (str, optional):
            The current document name.

            This is required when srcsets are not already cached.

    Returns:
        dict:
        The mapping of descriptors to URLs.
    """
    try:
        norm_srcsets = node.attributes['_srcsets']
    except KeyError:
        assert docname

        srcset = node.attributes.get('sources')
        norm_srcsets = OrderedDict()

        if srcset:
            norm_srcsets['1x'] = node.attributes['uri']

            for source in re.split(r',|\n+', srcset):
                source = source.strip()

                if source:
                    descriptor, url = source.split(' ', 1)
                    norm_srcsets[descriptor.strip()] = env.relfn2path(
                        search_image_for_language(url.strip(), env),
                        docname)[0]

        node.attributes['_srcsets'] = norm_srcsets

    return norm_srcsets


def _visit_image_html(
    self,
    node   # type: nodes.image
):  # type: (...) -> None
    """Process an Image node.

    This will update the HTML of the image node with a ``srcsets=`` attribute
    if srcsets are needed.

    Args:
        node (docutils.nodes.image):
            The image node to process.
    """
    # Use the default logic to build the image tag, since it's non-trivial.
    type(self).visit_image(self, node)

    builder = self.builder
    env = builder.env
    images = env.images
    base_images_path = builder.imgpath

    srcsets = _get_srcsets(node=node,
                           env=env)

    if srcsets:
        last_tag = self.body[-1]
        assert last_tag.startswith('<img ')

        new_attrs: List[str] = [
            'srcset="%s"' % ', '.join(
                '%s %s' % (
                    posixpath.join(base_images_path,
                                   urllib_quote(images[url][1])),
                    source)
                for source, url in srcsets.items()
            ),
        ]

        # Set a width attribute.
        m = WIDTH_ATTR_RE.search(last_tag)

        if not m and 'width' in node:
            new_attrs.append('width="%s"' % node['width'])

        # Set a height attribute.
        m = HEIGHT_ATTR_RE.search(last_tag)

        if not m and 'height' in node:
            new_attrs.append('height="%s"' % node['height'])

        # Remove dimensions from the style="" attribute.
        m = STYLE_ATTR_RE.search(last_tag)

        if m:
            styles = m.group('style').split(';')
            new_styles: List[str] = [
                style
                for style in styles
                if (style and
                    not style.lstrip().startswith(('width:', 'height:')))
            ]

            if new_styles:
                start, stop = m.span('style')
                last_tag = '%s%s%s' % (last_tag[:start],
                                       ';'.join(new_styles),
                                       last_tag[stop:])
            else:
                start, stop = m.span()
                last_tag = '%s%s' % (last_tag[:start],
                                     last_tag[stop:])

        self.body[-1] = '<img %s %s' % (
            ' '.join(new_attrs),
            last_tag[len('<img '):],
        )


def collect_srcsets(
    app,     # type: Sphinx
    doctree  # type: nodes.document
):  # type: (...) -> None
    """Collect all images referenced by image nodes or scanned in directories.

    This will collect any explicit values defined via our ``sources`` option
    for image directives. If ``sources`` is not specified, but there are files
    in the directory with ``@2x``, ``@3x``, ``@100w`` ``@100h``, etc.
    descriptors, those will be collected instead and associated with the image.

    Args:
        app (sphinx.application.Sphinx):
            The Sphinx application being run.

        doctree (docutils.nodes.document):
            The document tree being processed.
    """
    env = app.env
    images = env.images
    docname = env.docname

    if hasattr(doctree, 'findall'):
        # This is the modern way of finding nodes.
        findall = doctree.findall
    else:
        # This is pending deprecation in docutils.
        findall = doctree.traverse

    for node in findall(nodes.image):
        srcsets = _get_srcsets(node=node,
                               env=env,
                               docname=docname)

        if not srcsets:
            # NOTE: This will modify the contents of the cached srcsets.
            uri = node['uri']
            image_path = search_image_for_language(uri, env)
            base_filename, ext = os.path.splitext(image_path)
            candidates = glob('%s@*%s' % (base_filename, ext))

            if candidates:
                srcsets['1x'] = uri

                pattern = re.compile(r'%s@(\d+[xwh])%s'
                                     % (re.escape(base_filename),
                                        re.escape(ext)))

                for candidate in sorted(candidates):
                    m = pattern.match(candidate)

                    if m:
                        descriptor = m.group(1)

                        if descriptor not in srcsets:
                            srcsets[descriptor] = candidate

        for descriptor, image_path in srcsets.items():
            env.dependencies[docname].add(image_path)
            images.add_file(docname, image_path)


def collect_pages(
    app  # type: Sphinx
):  # type: (...) -> List
    """Collect srcset-specified images for use in HTML pages.

    This will go through the images referenced in a document for an HTML page
    and add any images found in srcsets to the list of images to collect for
    the page.

    Args:
        app (sphinx.application.Sphinx):
            The Sphinx application to register roles and configuration on.

    Returns:
        list:
        An empty list (indicating no additional HTML pages are collected).
    """
    app.builder.images.update({
        full_path: filename
        for full_path, (docnames, filename) in app.env.images.items()
    })

    return []


def setup(
    app  # type: Sphinx
):  # type: (...) -> Dict
    """Set up the Sphinx extension.

    This listens for the events needed to collect and bundle images for
    srcsets, and update the resulting HTML to specify them.

    Args:
        app (sphinx.application.Sphinx):
            The Sphinx application being run.
    """
    Image.option_spec['sources'] = directives.unchanged

    app.add_node(nodes.image,
                 html=(_visit_image_html, None),
                 override=True)

    app.connect('doctree-read', collect_srcsets)
    app.connect('html-collect-pages', collect_pages)

    return {
        'version': VERSION,
    }



# -*- coding: utf-8 -*-
#
# beanbag-docutils documentation build configuration file, created by
# sphinx-quickstart on Fri Jul 15 14:17:09 2016.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(__file__, '..', '..')))

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'beanbag_docutils.sphinx.ext.autodoc_utils',
    'beanbag_docutils.sphinx.ext.http_role',
    'beanbag_docutils.sphinx.ext.extlinks',
    'beanbag_docutils.sphinx.ext.ref_utils',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'beanbag-docutils'
copyright = '2021, Beanbag, Inc.'
author = 'Beanbag, Inc.'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = ''

# The full version, including alpha/beta/rc tags.
release = ''

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# If true, '()' will be appended to :func: etc. cross-reference text.
#
add_function_parentheses = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'classic'
html_style = 'classic.css'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    'django': ('https://docs.djangoproject.com/en/3.1/',
               'https://docs.djangoproject.com/en/3.1/_objects/'),
    'pyhton': ('https://docs.python.org/', None),
}

extlinks = {
    'pypi': ('https://pypi.org/project/%s/', '%s'),
}


autosummary_generate = True
napolean_beanbag_docstring = True
napolean_google_docstring = False
napolean_numpy_docstring = False



"""Unit tests for beanbag_docutils.sphinx.ext.autodoc_utils."""

from sphinx import version_info as sphinx_version_info

from beanbag_docutils.sphinx.ext import autodoc_utils
from beanbag_docutils.sphinx.ext.autodoc_utils import BeanbagDocstring
from beanbag_docutils.sphinx.ext.tests.testcase import SphinxExtTestCase


# A series of modules used to test autodoc exclusions.
class IgnoredModule(object):
    pass


class DeprecatedModule(object):
    pass


class DocClass1(object):
    def foo(self):
        pass

    def bar(self):
        pass


class DocClass2(object):
    def foo(self):
        pass


class AutoDocExcludesTests(SphinxExtTestCase):
    """Unit tests for autodoc_excludes setting."""

    config = {
        'autoclass_content': 'class',
        'autodoc_default_options': {
            'members': None,
            'member-order': 'bysource',
            'special-members': None,
            'undoc-members': None,
        },
    }

    extensions = [
        autodoc_utils.__name__,
    ]

    def test_with_autodoc_excludes(self):
        """Testing autodoc_excludes with __autodoc_excludes__ in module"""
        rendered = self.render_doc('.. automodule:: %s\n' % __name__)

        self.assertIn('AutoDocExcludesTests', rendered)
        self.assertNotIn('IgnoredModule', rendered)

    def test_with_wildcard_exclude(self):
        """Testing autodoc_excludes with wildcard (*) exclude list"""
        doc = '\n'.join([
            '.. autoclass:: %s.%s' % (DocClass1.__module__,
                                      DocClass2.__name__),
            '',
            '.. autoclass:: %s.%s' % (DocClass1.__module__,
                                      DocClass2.__name__),
        ])

        rendered = self.render_doc(
            doc,
            config={
                'autodoc_excludes': {
                    '*': ['__dict__', 'foo'],
                },
            }
        )

        self.assertNotIn('__dict__', rendered)
        self.assertNotIn('foo', rendered)
        self.assertIn('__module__', rendered)

    def test_with_wildcard_exclude_with_defaults(self):
        """Testing autodoc_excludes with wildcard (*) exclude list with
        defaults
        """
        doc = '\n'.join([
            '.. autoclass:: %s.%s' % (DocClass1.__module__,
                                      DocClass2.__name__),
            '',
            '.. autoclass:: %s.%s' % (DocClass1.__module__,
                                      DocClass2.__name__),
        ])

        rendered = self.render_doc(doc)

        self.assertIn('foo', rendered)

    def test_with_wildcard_exclude_added_to_defaults(self):
        """Testing autodoc_excludes with wildcard (*) exclude list added to
        defaults
        """
        doc = '\n'.join([
            '.. autoclass:: %s.%s' % (DocClass1.__module__,
                                      DocClass1.__name__),
            '',
            '.. autoclass:: %s.%s' % (DocClass2.__module__,
                                      DocClass2.__name__),
        ])

        rendered = self.render_doc(
            doc,
            config={
                'autodoc_excludes': {
                    '__defaults__': True,
                    '*': ['foo'],
                },
            }
        )

        self.assertNotIn('__dict__', rendered)
        self.assertNotIn('foo', rendered)
        self.assertNotIn('__module__', rendered)

    def test_with_class_attr_exclude(self):
        """Testing autodoc_excludes with class exclude list"""
        docclass1_path = '%s.%s' % (DocClass1.__module__, DocClass1.__name__)

        doc = '\n'.join([
            '.. autoclass:: %s' % docclass1_path,
            '',
            '.. autoclass:: %s.%s' % (DocClass2.__module__,
                                      DocClass2.__name__),
        ])

        # Now exclude them.
        rendered = self.render_doc(
            doc,
            config={
                'autodoc_excludes': {
                    'class': ['__dict__', 'foo'],
                },
            }
        )

        self.assertIn('DocClass1.__module__', rendered)
        self.assertIn('DocClass2.__module__', rendered)
        self.assertIn('DocClass1.bar', rendered)
        self.assertNotIn('DocClass1.__dict__', rendered)
        self.assertNotIn('DocClass1.foo', rendered)
        self.assertNotIn('DocClass2.__dict__', rendered)
        self.assertNotIn('DocClass2.foo', rendered)

    def test_with_class_attr_exclude_with_defaults(self):
        """Testing autodoc_excludes with class exclude list with defaults"""
        docclass1_path = '%s.%s' % (DocClass1.__module__, DocClass1.__name__)

        doc = '\n'.join([
            '.. autoclass:: %s' % docclass1_path,
            '',
            '.. autoclass:: %s.%s' % (DocClass2.__module__,
                                      DocClass2.__name__),
        ])

        # Make sure the attributes on these classes are here by default.
        rendered = self.render_doc(doc)

        self.assertIn('DocClass1.foo', rendered)
        self.assertIn('DocClass1.bar', rendered)
        self.assertIn('DocClass2.foo', rendered)
        self.assertNotIn('DocClass1.__dict__', rendered)
        self.assertNotIn('DocClass1.__module__', rendered)
        self.assertNotIn('DocClass2.__dict__', rendered)
        self.assertNotIn('DocClass2.__module__', rendered)

    def test_with_class_attr_exclude_and_added_to_defaults(self):
        """Testing autodoc_excludes with class exclude list and added with
        defaults
        """
        docclass1_path = '%s.%s' % (DocClass1.__module__, DocClass1.__name__)

        doc = '\n'.join([
            '.. autoclass:: %s' % docclass1_path,
            '',
            '.. autoclass:: %s.%s' % (DocClass1.__module__,
                                      DocClass2.__name__),
        ])

        # Now exclude them.
        rendered = self.render_doc(
            doc,
            config={
                'autodoc_excludes': {
                    '__defaults__': True,
                    'class': ['foo'],
                },
            }
        )

        self.assertIn('DocClass1.bar', rendered)
        self.assertNotIn('DocClass1.__module__', rendered)
        self.assertNotIn('DocClass1.__dict__', rendered)
        self.assertNotIn('DocClass1.foo', rendered)
        self.assertNotIn('DocClass2.__module__', rendered)
        self.assertNotIn('DocClass2.__dict__', rendered)
        self.assertNotIn('DocClass2.foo', rendered)

    def test_with_empty_excludes(self):
        """Testing autodoc_excludes with empty excludes"""
        doc = '\n'.join([
            '.. autoclass:: %s.%s' % (DocClass1.__module__,
                                      DocClass1.__name__),
            '',
            '.. autoclass:: %s.%s' % (DocClass2.__module__,
                                      DocClass2.__name__),
        ])

        # Make sure some our attributes are here by default.
        rendered = self.render_doc(
            doc,
            config={
                'autodoc_excludes': {},
            })

        self.assertIn('__dict__', rendered)
        self.assertIn('foo', rendered)
        self.assertIn('__module__', rendered)
        self.assertIn('DocClass1.__module__', rendered)

    def test_with_deprecated(self):
        """Testing autodoc_excludes with __deprecated__ in module"""
        rendered = self.render_doc('.. automodule:: %s\n' % __name__)

        self.assertIn('AutoDocExcludesTests', rendered)
        self.assertNotIn('DeprecatedModule', rendered)


class BeanbagDocstringTests(SphinxExtTestCase):
    """Unit tests for BeanbagDocstring."""

    extensions = [
        autodoc_utils.__name__,
    ]

    def test_args_section(self):
        """Testing Beanbag docstring with Args section"""
        rendered = self._render_docstring(
            'Args:\n'
            '   arg1 (str):\n'
            '       Description of arg1.\n'
            '\n'
            '   arg2 (foo.bar.abc\n'
            '         .def.ghi,\n'
            '         optional):\n'
            '       Description of arg2.\n'
            '\n'
            '   *args (tuple):\n'
            '       Description of args.\n'
            '\n'
            '   **kwargs (dict):\n'
            '       Description of kwargs.\n'
        )

        if sphinx_version_info[:2] >= (7, 2):
            self.assertEqual(
                rendered,
                ':param arg1: Description of arg1.\n'
                ':type arg1: :py:class:`str`\n'
                ':param arg2: Description of arg2.\n'
                ':type arg2: :py:class:`foo.bar.abc.def.ghi`, *optional*\n'
                ':param \\*args: Description of args.\n'
                ':type \\*args: :py:class:`tuple`\n'
                ':param \\*\\*kwargs: Description of kwargs.\n'
                ':type \\*\\*kwargs: :py:class:`dict`\n')
        else:
            self.assertEqual(
                rendered,
                ':param arg1: Description of arg1.\n'
                ':type arg1: :class:`str`\n'
                ':param arg2: Description of arg2.\n'
                ':type arg2: :class:`foo.bar.abc.def.ghi`, *optional*\n'
                ':param \\*args: Description of args.\n'
                ':type \\*args: :class:`tuple`\n'
                ':param \\*\\*kwargs: Description of kwargs.\n'
                ':type \\*\\*kwargs: :class:`dict`\n')

    def test_context_section_with_description(self):
        """Testing Beanbag docstring with Context section with description"""
        self.assertEqual(
            self._render_docstring(
                'Context:\n'
                '    Description of the context.\n'
            ),
            ':Context: Description of the context.\n'
        )

    def test_context_section_with_type(self):
        """Testing Beanbag docstring with Context section with type"""
        self.assertEqual(
            self._render_docstring(
                'Context:\n'
                '    dict\n'
            ),
            ':Context: dict\n'
        )

    def test_context_section_with_type_and_description(self):
        """Testing Beanbag docstring with Context section with type and
        description
        """
        rendered = self._render_docstring(
            'Context:\n'
            '    dict:\n'
            '    Description of the context.\n'
        )

        if sphinx_version_info[:2] >= (7, 2):
            self.assertEqual(
                rendered,
                ':Context: :py:class:`dict` -- Description of the context.\n')
        else:
            self.assertEqual(
                rendered,
                ':Context: :class:`dict` -- Description of the context.\n')

    def test_deprecated_section_with_version(self):
        """Testing Beanbag docstring with Deprecated section with version"""
        self.assertEqual(
            self._render_docstring(
                'Deprecated:\n'
                '    2.0\n'
            ),
            '.. deprecated:: 2.0\n\n'
        )

    def test_deprecated_section_with_version_and_description(self):
        """Testing Beanbag docstring with Deprecated section with version and
        description
        """
        self.assertEqual(
            self._render_docstring(
                'Deprecated:\n'
                '    2.0:\n'
                '    Description of the deprecation.\n'
            ),
            (
                '.. deprecated:: 2.0\n'
                '\n'
                '   Description of the deprecation.\n'
            )
        )

    def test_deprecated_section_with_version_and_lists(self):
        """Testing Beanbag docstring with Deprecated section with lists"""
        self.assertEqual(
            self._render_docstring(
                'Deprecated:\n'
                '    2.0:\n'
                '    * This is item one and it\n'
                '      wraps across multiple\n'
                '      lines\n'
                '\n'
                '    * And this is item 2.\n'
                '      Still multiple lines.\n'
            ),
            (
                '.. deprecated:: 2.0\n'
                '\n'
                '   * This is item one and it\n'
                '     wraps across multiple\n'
                '     lines\n'
                '   \n'
                '   * And this is item 2.\n'
                '     Still multiple lines.\n'
            )
        )

    def test_keys(self):
        """Testing Beanbag docstring with Keys section"""
        rendered = self._render_docstring(
            'Keys:\n'
            '    key1 (str):\n'
            '        Description 1\n'
            '\n'
            '    key2 (dict):\n'
            '        Description 2\n'
            '\n'
            '    key3 (int, optional):\n'
            '        Description 3\n'
        )

        if sphinx_version_info[:2] >= (7, 2):
            self.assertEqual(
                rendered,
                ':Keys: * **key1** (:py:class:`str`) -- Description 1\n'
                '       * **key2** (:py:class:`dict`) -- Description 2\n'
                '       * **key3** (:py:class:`int`, *optional*)'
                ' -- Description 3\n')
        else:
            self.assertEqual(
                rendered,
                ':Keys: * **key1** (:class:`str`) -- Description 1\n'
                '       * **key2** (:class:`dict`) -- Description 2\n'
                '       * **key3** (:class:`int`, *optional*)'
                ' -- Description 3\n')

    def test_model_attributes(self):
        """Testing Beanbag docstring with Model Attributes section"""
        rendered = self._render_docstring(
            'Model Attributes:\n'
            '    attr1 (dict):\n'
            '        Description of attr1\n'
            '\n'
            '    attr2 (foo.bar\n'
            '           .baz):\n'
            '        Description of attr2\n'
        )

        if sphinx_version_info[:2] >= (7, 2):
            self.assertEqual(
                rendered,
                ':Model Attributes: * **attr1** (:py:class:`dict`)'
                ' -- Description of attr1\n'
                '                   * **attr2** (:py:class:`foo.bar.baz`)'
                ' -- Description of attr2\n')
        else:
            self.assertEqual(
                rendered,
                ':Model Attributes: * **attr1** (:class:`dict`)'
                ' -- Description of attr1\n'
                '                   * **attr2** (:class:`foo.bar.baz`)'
                ' -- Description of attr2\n')


    def test_option_args(self):
        """Testing Beanbag docstring with Option Args section"""
        rendered = self._render_docstring(
            'Option Args:\n'
            '    attr1 (dict):\n'
            '        Description of attr1\n'
            '\n'
            '    attr2 (foo.bar\n'
            '           .baz,\n'
            '           optional):\n'
            '        Description of attr2\n'
        )

        if sphinx_version_info[:2] >= (7, 2):
            self.assertEqual(
                rendered,
                ':Option Args: * **attr1** (:py:class:`dict`)'
                ' -- Description of attr1\n'
                '              * **attr2** (:py:class:`foo.bar.baz`,'
                ' *optional*) -- Description of attr2\n')
        else:
            self.assertEqual(
                rendered,
                ':Option Args: * **attr1** (:class:`dict`)'
                ' -- Description of attr1\n'
                '              * **attr2** (:class:`foo.bar.baz`, *optional*)'
                ' -- Description of attr2\n')

    def test_tuple(self):
        """Testing Beanbag docstring with Tuple section"""
        rendered = self._render_docstring(
            'Tuple:\n'
            '    0 (str):\n'
            '        Description 1\n'
            '\n'
            '    1 (dict):\n'
            '        Description 2\n'
            '\n'
            '    2 (int, optional):\n'
            '        Description 3\n'
        )

        if sphinx_version_info[:2] >= (7, 2):
            self.assertEqual(
                rendered,
                ':Tuple: * **0** (:py:class:`str`) -- Description 1\n'
                '        * **1** (:py:class:`dict`) -- Description 2\n'
                '        * **2** (:py:class:`int`, *optional*) -- '
                'Description 3\n')
        else:
            self.assertEqual(
                rendered,
                ':Tuple: * **0** (:class:`str`) -- Description 1\n'
                '        * **1** (:class:`dict`) -- Description 2\n'
                '        * **2** (:class:`int`, *optional*) -- '
                'Description 3\n')

    def test_type_section(self):
        """Testing Beanbag docstring with Type section"""
        rendered = self._render_docstring(
            'Type:\n'
            '    dict\n'
        )

        if sphinx_version_info[:2] >= (7, 2):
            self.assertEqual(rendered, ':Type: :py:class:`dict`\n')
        else:
            self.assertEqual(rendered, ':Type: :class:`dict`\n')

    def test_type_section_with_description(self):
        """Testing Beanbag docstring with Type section with description"""
        rendered = self._render_docstring(
            'Type:\n'
            '    dict:\n'
            '    Description.\n'
        )

        if sphinx_version_info[:2] >= (7, 2):
            self.assertEqual(rendered,
                             ':Type: :py:class:`dict` -- Description.\n')
        else:
            self.assertEqual(rendered,
                             ':Type: :class:`dict` -- Description.\n')

    def test_version_added_section_with_version(self):
        """Testing Beanbag docstring with Version Added section with version"""
        self.assertEqual(
            self._render_docstring(
                'Version Added:\n'
                '    2.0\n'
            ),
            (
                '.. versionadded:: 2.0\n'
                '\n'
            )
        )

    def test_version_added_section_with_version_and_description(self):
        """Testing Beanbag docstring with Version Added section with version
        and description
        """
        self.assertEqual(
            self._render_docstring(
                'Version Added:\n'
                '    2.0:\n'
                '    Description of the addition.\n'
            ),
            (
                '.. versionadded:: 2.0\n'
                '\n'
                '   Description of the addition.\n'
            )
        )

    def test_version_added_section_with_lists(self):
        """Testing Beanbag docstring with Version Added section with lists"""
        self.assertEqual(
            self._render_docstring(
                'Version Added:\n'
                '    2.0:\n'
                '    * This is item one and it\n'
                '      wraps across multiple\n'
                '      lines\n'
                '\n'
                '    * And this is item 2.\n'
                '      Still multiple lines.\n'
            ),
            (
                '.. versionadded:: 2.0\n'
                '\n'
                '   * This is item one and it\n'
                '     wraps across multiple\n'
                '     lines\n'
                '   \n'
                '   * And this is item 2.\n'
                '     Still multiple lines.\n'
            )
        )

    def test_version_changed_section(self):
        """Testing Beanbag docstring with Version Changed section"""
        self.assertEqual(
            self._render_docstring(
                'Version Changed:\n'
                '    2.0:\n'
                '    Description of the change.\n'
            ),
            (
                '.. versionchanged:: 2.0\n'
                '\n'
                '   Description of the change.\n'
            )
        )

    def test_version_changed_section_with_lists(self):
        """Testing Beanbag docstring with Version Changed section with lists"""
        self.assertEqual(
            self._render_docstring(
                'Version Changed:\n'
                '    2.0:\n'
                '    * This is item one and it\n'
                '      wraps across multiple\n'
                '      lines\n'
                '\n'
                '    * And this is item 2.\n'
                '      Still multiple lines.\n'
            ),
            (
                '.. versionchanged:: 2.0\n'
                '\n'
                '   * This is item one and it\n'
                '     wraps across multiple\n'
                '     lines\n'
                '   \n'
                '   * And this is item 2.\n'
                '     Still multiple lines.\n'
            )
        )

    def _render_docstring(self, content):
        """Render a Beanbag docstring to ReST.

        Args:
            content (unicode):
                The docstring content to render.

        Returns:
            unicode:
            The resulting ReStructuredText.
        """
        with self.with_sphinx_env() as ctx:
            return str(BeanbagDocstring(content, config=ctx['config']))


__autodoc_excludes__ = ['IgnoredModule']
__deprecated__ = ['DeprecatedModule']



"""Unit tests for beanbag_docutils.sphinx.ext.json_writer.

Version Added:
    2.2
"""

import json
import os

from beanbag_docutils.sphinx.ext import json_writer
from beanbag_docutils.sphinx.ext.tests.testcase import SphinxExtTestCase


class JsonWriterTests(SphinxExtTestCase):
    extensions = [
        json_writer.__name__,
    ]

    def test_anchors_html(self):
        """Testing json_writer with anchors_html"""
        data = json.loads(self.render_doc(
            '==========\n'
            'Page Title\n'
            '==========\n'
            '\n'
            '.. _section-1::\n'
            'Section 1\n'
            '=========\n'
            '\n'
            '.. _section-1-2::\n'
            'Section 1.2\n'
            '---------\n'
            '\n'
            '.. _section-2::\n'
            'Section 2\n'
            '=========\n',
            builder_name='json'))

        self.assertEqual(
            data.get('anchors_html'),
            '<ul>\n'
            '<li><a class="reference internal" href="#section-1">Section 1</a>'
            '<ul>\n'
            '<li><a class="reference internal" href="#section-1-2">Section 1.2'
            '</a></li>\n'
            '</ul>\n'
            '</li>\n'
            '<li><a class="reference internal" href="#section-2">Section 2</a>'
            '</li>\n'
            '</ul>\n'
        )

    def test_anchors_html_with_no_sections(self):
        """Testing json_writer with anchors_html with no sections"""
        data = json.loads(self.render_doc(
            '==========\n'
            'Page Title\n'
            '==========\n',
            builder_name='json'))

        self.assertIn('anchors_html', data)
        self.assertIsNone(data.get('anchors_html'))

    def test_toc(self):
        """Testing json_writer with toc"""
        files = {
            'contents.rst': '',
            'index.rst': (
                '========\n'
                'Top Page\n'
                '========\n'
                '\n'
                '.. toctree::\n'
                '\n'
                '   page1\n'
                '   page2\n'
            ),
            'page1.rst': (
                '======\n'
                'Page 1\n'
                '======\n'
                '\n'
                '.. toctree::\n'
                '\n'
                '   page1.1\n'
            ),
            'page1.1.rst': (
                '========\n'
                'Page 1.1\n'
                '========\n'
            ),
            'page2.rst': (
                '======\n'
                'Page 2\n'
                '======\n'
            ),
        }

        with self.rendered_docs(files=files,
                                builder_name='json') as build_dir:
            with open(os.path.join(build_dir, 'globalcontext.json'),
                      'r') as fp:
                global_context = json.load(fp)

            self.assertIn('toc', global_context)
            self.assertEqual(global_context['toc'], [
                {
                    'docname': 'index',
                    'title': 'Top Page',
                    'items': [
                        {
                            'docname': 'page1',
                            'title': 'Page 1',
                            'items': [
                                {
                                    'docname': 'page1.1',
                                    'title': 'Page 1.1',
                                },
                            ],
                        },
                        {
                            'docname': 'page2',
                            'title': 'Page 2',
                        },
                    ],
                },
            ])

    def test_toc_with_no_children(self):
        """Testing json_writer with toc and no children"""
        files = {
            'contents.rst': '',
            'index.rst': (
                '========\n'
                'Top Page\n'
                '========\n'
                '\n'
                '.. toctree::\n'
                '\n'
                '   page1\n'
                '   page2\n'
            ),
        }

        with self.rendered_docs(files=files,
                                builder_name='json') as build_dir:
            with open(os.path.join(build_dir, 'globalcontext.json'),
                      'r') as fp:
                global_context = json.load(fp)

            self.assertIn('toc', global_context)
            self.assertEqual(global_context['toc'], [
                {
                    'docname': 'index',
                    'title': 'Top Page',
                },
            ])

    def test_toc_with_doc_root_page(self):
        """Testing json_writer with toc and doc_root page (no index.rst)"""
        files = {
            'contents.rst': (
                '========\n'
                'Top Page\n'
                '========\n'
                '\n'
                '.. toctree::\n'
                '\n'
                '   page1\n'
                '   page2\n'
            ),
        }

        with self.rendered_docs(files=files,
                                builder_name='json') as build_dir:
            with open(os.path.join(build_dir, 'globalcontext.json'),
                      'r') as fp:
                global_context = json.load(fp)

            self.assertIn('toc', global_context)
            self.assertEqual(global_context['toc'], [
                {
                    'docname': 'contents',
                    'title': 'Top Page',
                },
            ])



"""Sphinx extension for extracting metadata for the page.

.. versionadded:: 2.2


This extension extracts any :rst:dir:`meta` information (useful for content
like page descriptions for social media) and extracts their information into
the document's top-level metadata.

This metadata can then be used when building sites or tools that process the
page's compiled documentation (when not simply relying on the generated HTML).

The metadata will be available in the document's ``meta`` information, keyed
off by the name used in the :rst:dir:`meta` directive, the content, and any
other attributes provided.

If multiple pieces of metadata exist for the same key, the information will be
assembled into a list under that key. Otherwise, the information will be
assigned directly to the key.


Setup
=====

To use this, you just need to add the extension in :file:`conf.py`::

    extensions = [
        ...
        'beanbag_docutils.sphinx.ext.metadata',
        ...
    ]
"""

from typing import Any, Dict, TYPE_CHECKING

import six
from docutils import nodes, __version__ as DOCUTILS_VERSION
from sphinx.errors import ExtensionError

from beanbag_docutils import VERSION

if TYPE_CHECKING:
    from sphinx.application import Sphinx


def _on_doctree_read(
    app,     # type: Sphinx
    doctree  # type: nodes.document
):  # type: (...) -> None
    """Extract any doc-defined metadata into the compiled page information.

    This will go through any :rst:dir:`meta` nodes, pulling out attributes
    and storing them in the metadata for the document.

    Version Added:
        2.2

    Args:
        app (sphinx.application.Sphinx):
            The Sphinx application being run.

        doctree (docutils.nodes.document):
            The document tree being processed.
    """
    metadata = {}  # type: Dict[str, Any]
    env = app.env

    if hasattr(doctree, 'findall'):
        # This is the modern way of finding nodes.
        findall = doctree.findall
    else:
        # This is pending deprecation in docutils.
        findall = doctree.traverse
        assert False

    for node in findall(nodes.meta):
        if not node.hasattr('name'):
            continue

        name = node['name']
        meta_attrs = dict(node.non_default_attributes())
        del meta_attrs['name']

        metadata.setdefault(name, []).append(meta_attrs)

    # Register all the new metadata.
    docname = env.docname
    doc_metadata = env.metadata[docname]

    for key, values in six.iteritems(metadata):
        if len(values) == 1:
            doc_metadata[key] = values[0]
        else:
            doc_metadata[key] = values


def setup(
    app  # type: Sphinx
):  # type: (...) -> Dict
    """Set up the Sphinx extension.

    This listens for the event needed to process the document tree for
    metadata extraction.

    Version Added:
        2.2

    Args:
        app (sphinx.application.Sphinx):
            The Sphinx application being run.
    """
    if not hasattr(nodes, 'meta'):
        raise ExtensionError('beanbag_docutils.sphinx.ext does not support '
                             'docutils %s'
                             % DOCUTILS_VERSION)

    app.connect('doctree-read', _on_doctree_read)

    return {
        'version': VERSION,
    }



"""Sphinx extension to define external links that support anchors.

Sphinx comes bundled with a :py:mod:`sphinx.ext.extlinks` extension, which
allows a :file:`conf.py` to define roles for external links. These don't
support anchors, however, making it impossible to link to properly link in some
cases.

This is a wrapper around that module that looks for anchors and appends them to
the resulting URL.


Setup
=====

This extension works identically to :py:mod:`sphinx.ext.extlinks` and contains
the same configuration. To use it, configure external links the way you would
for that extension, but add ours instead to :file:`conf.py`::

    extensions = [
        ...
        'beanbag_docutils.sphinx.ext.extlinks',
        ...
    ]
"""

from sphinx import version_info as sphinx_version_info
from sphinx.ext.extlinks import make_link_role


class ExternalLink(object):
    """Wraps a URL and formats references to allow for using anchors.

    This will work like a string, from the point of view of
    :py:mod:`sphinx.ext.extlinks`. It takes the URL for the external link and
    intercepts any string formatting, pulling out the anchor and appending it
    to the final result.
    """

    def __init__(self, base_url):
        """Initialize the class.

        Args:
            base_url (unicode):
                The URL to wrap. This must contain a ``%s``.
        """
        self.base_url = base_url

    def __mod__(self, ref):
        """Return a URL based on the stored string format and the reference.

        Args:
            ref (unicode):
                The reference to place into the URL. This may contain an
                anchor starting with ``#``.

        Returns:
            unicode:
            The formatted URL.
        """
        parts = ref.split('#', 1)
        url = self.base_url % parts[0]

        if len(parts) == 2:
            url = '%s#%s' % (url, parts[1])

        return url

    def __add__(self, s):
        """Return the concatenated string for the base URL and another string.

        Args:
            s (unicode):
                A string to concatenate onto this base URL.

        Returns:
            unicode:
            The concatenated string.
        """
        return self.base_url + s


def setup_link_roles(app):
    """Register roles for each external link that's been defined.

    Args:
        app (sphinx.application.Sphinx):
            The Sphinx application.
    """
    if sphinx_version_info[:2] >= (4, 0):
        for name, (base_url, prefix) in app.config.extlinks.items():
            app.add_role(name, make_link_role(name, ExternalLink(base_url),
                                              prefix))
    else:
        for name, (base_url, prefix) in app.config.extlinks.items():
            app.add_role(name, make_link_role(ExternalLink(base_url), prefix))


def setup(app):
    """Set up the Sphinx extension.

    This registers the configuration for external links and adds the roles
    for each when the builder initializes.

    Args:
        app (sphinx.application.Sphinx):
            The Sphinx application.

    Returns:
        dict:
        Information about the extension. This is in the same format as what
        :py:func:`sphinx.ext.extlinks.setup` returns.
    """
    app.add_config_value('extlinks', {}, 'env')
    app.connect('builder-inited', setup_link_roles)

    return {
        'version': '1.0',
        'parallel_read_safe': True,
    }



"""Sphinx extension to collect additional files in the build directory.

This is used to copy files (indicated by glob patterns) from the source
directory into the destination build directory. Each destination file will be
in the same relative place in the tree.

This is useful when you have non-ReST/image files that you want part of your
built set of files, perhaps containing metadata or packaging that you want to
ship along with the documentation.


Setup
=====

To use this, you just need to add the extension in :file:`conf.py`::

    extensions = [
        ...
        'beanbag_docutils.sphinx.ext.collect_files',
        ...
    ]

And then configure ``collect_file_patterns`` to be a list of
filenames/glob patterns.


Configuration
=============

``collect_file_patterns``
    List of filenames or glob patterns to include from the source directory
    into the build directory.
"""

import os
import shutil
from fnmatch import fnmatch


def collect_files(app, env):
    """Collect configured files and put them into the build directory.

    Args:
        app (sphinx.application.Sphinx):
            The Sphinx application to register roles and configuration on.

        env (sphinx.environment.BuildEnvironment, unused):
            The build environment for the generated docs.
    """
    collect_patterns = app.config['collect_file_patterns']
    src_dir = app.builder.srcdir
    out_dir = app.builder.outdir

    for root, dirs, files in os.walk(src_dir):
        # Make sure we don't recurse into the build directory.
        if root == src_dir:
            try:
                dirs.remove('_build')
            except ValueError:
                pass

        for filename in files:
            for pattern in collect_patterns:
                if fnmatch(filename, pattern):
                    shutil.copy(os.path.join(root, filename),
                                os.path.join(out_dir,
                                             os.path.relpath(root, src_dir),
                                             filename))
                    break


def setup(app):
    """Set up the Sphinx extension.

    This listens for the events needed to collect files.

    Args:
        app (sphinx.application.Sphinx):
            The Sphinx application to listen to events on.

    Returns:
        dict:
        Information about the extension.
    """
    app.add_config_value('collect_file_patterns', {}, True)
    app.connect('env-updated', collect_files)

    return {
        'version': '1.0',
        'parallel_read_safe': True,
    }



"""Sphinx extension offering an enhanced JSON document builder.

When enabled, the JSON builder provided by
:pypi:`sphinxcontrib-serializinghtml` will be augmented with:

1. A ``toc`` key in :file:`globalcontext.json` containing a structured
   Table of Contents for the site.

2. An ``anchors_html`` in the page's context containing HTML-rendered anchors
   for the page, without the page title being included.


Setup
=====

To use this, add the extension in :file:`conf.py`::

    extensions = [
        ...
        'beanbag_docutils.sphinx.ext.json_writer',
        ...
    ]

This will automatically enable the :py:mod:`sphinxcontrib.serializinghtml`
extension and set the appropriate overrides.

You can then build the JSON documentation using the ``json`` builder.


Table of Contents Structure
===========================

The ``toc`` key in :file:`globalcontext.json` contains a list of dictionaries
containing:

Keys:
   docname (str):
       The name of the document.

   title (str):
       The document title.

   items (list, optional):
       The list of child documents under this page in the tree. This uses
       this same structure.


This will start with the ``index.rst`` at the top of the project, if it exists.
If it does not, it will fall back to the configured Sphinx root document
(configured by ``root_doc``).


Anchors Structure
=================

Anchors are stored in a ``.fjson`` page's ``anchors_html`` key. This will be
pre-rendered HTML, following the form of the normal ``toc`` key. For example:

.. code-block:: html

    <ul>
     <li><a class="reference internal" href="#section1">Section 1</a></li>
     <li><a class="reference internal" href="#section2">Section 2</a><ul>
      <li><a class="reference internal" href="#section2.1">Section 2.1</a></li>
      <li><a class="reference internal" href="#section2.2">Section 2.2</a></li>
      ...
     </ul></li>
     ...
    </ul>

"""

from typing import Any, Dict, List, TYPE_CHECKING

from sphinx.environment.adapters.toctree import TocTree

from beanbag_docutils import VERSION

try:
    from sphinx.builders.html import JSONHTMLBuilder
    has_native_json_builder = True
except ImportError:
    from sphinxcontrib.serializinghtml import JSONHTMLBuilder
    has_native_json_builder = False

if TYPE_CHECKING:
    from sphinx.application import Sphinx


class JSONBuilder(JSONHTMLBuilder):
    """Sphinx builder for JSON files.

    This specializes :py:mod:`sphinxcontrib.serializinghtml.JSONHTMLBuilder`,
    adding additional state for a structured Table of Contents, available
    in :file:`glboalcontext.json`.

    Version Added:
        2.2
    """

    def handle_finish(self):  # type: (...) -> None
        """Handle finishing the build for all the docs."""
        env = self.env
        toc = None

        for name in ('index', getattr(env.config, 'root_doc', 'contents')):
            if name and name in env.titles:
                toc = self._build_toc([name])
                break

        self.globalcontext['toc'] = toc

        super(JSONBuilder, self).handle_finish()

    def _build_toc(
        self,
        docnames,  # type: List[str]
    ):  # type: (...) -> List[Dict[str, Any]]
        """Build the Table of Contents for a given level

        This will iterate through ``docnames`` and produce entries for each,
        recursively building for any child documents.

        Args:
            docnames (list of str):
                The list of document names on this level.

        Returns:
            list of dict:
            The list of Table of Contents entries for this level.
        """
        env = self.env
        toc = []  # type: List[Dict[str, Any]]

        for docname in docnames:
            toc_info = {
                'docname': docname,
                'title': str(env.titles[docname].children[0]),
            }  # type: Dict[str, Any]

            children = env.toctree_includes.get(docname, [])

            if children:
                toc_info['items'] = self._build_toc(children)

            toc.append(toc_info)

        return toc


def _on_html_page_context(
    app,           # type: Sphinx
    pagename,      # type: str
    templatename,  # type: str
    ctx,           # type: Dict
    event_arg      # type: Any
):  # type: (...) -> None
    """Gather information for the page.

    This will generate the ``anchors_html`` for a page and set it in the
    context. This may be ``None`` if a Table of Contents is not present.

    Version Added:
        2.2

    Args:
        app (sphinx.application.Sphinx):
            The current Sphinx application.

        pagename (str):
            The name of the page.

        templatename (str, unused):
            The template being used to render the page.

        ctx (dict):
            The context to populate.

        event_arg (object, unused):
            An unused event argument from the caller.
    """
    builder = app.builder
    toc = TocTree(app.env).get_toc_for(pagename, builder)
    anchors_html = None

    children = toc.children

    if len(children) > 0:
        toc_top = children[0]
        anchors_node = None

        if len(toc_top.children) > 1:
            anchors_node = toc_top.children[1]
        elif len(children) > 1:
            anchors_node = children[1]

        if anchors_node is not None:
            anchors_html = builder.render_partial(anchors_node)['fragment']

    ctx['anchors_html'] = anchors_html


def setup(app):
    """Set up the Sphinx extension.

    This sets up the configuration and event handlers needed for the JSON
    writer extension.

    Version Added:
        2.2

    Args:
        app (sphinx.application.Sphinx):
            The Sphinx application to register configuration and listen to
            events on.
    """
    if not has_native_json_builder:
        app.setup_extension('sphinxcontrib.serializinghtml')

    app.add_builder(JSONBuilder, override=True)

    app.connect('html-page-context', _on_html_page_context)

    return {
        'version': VERSION,
    }



"""Unit tests for beanbag_docutils.sphinx.ext.ref_utils."""

from beanbag_docutils.sphinx.ext.tests.testcase import SphinxExtTestCase


class RefUtilsTests(SphinxExtTestCase):
    """Unit tests for reference role utilities."""

    extensions = [
        'beanbag_docutils.sphinx.ext.ref_utils',
    ]

    def test_ref_break_for_javascript(self):
        """Testing ref_utils reference line break for JavaScript references"""
        rendered = self.render_doc(
            '.. js:function:: abc.def.myfunc\n'
            '\n'
            '   Description.\n'
            '\n'
            'Link to :js:class:`abc.def\n'
            '.myfunc`.'
        )

        self.assertIn(
            '<a class="reference internal" href="#abc.def.myfunc"'
            ' title="abc.def.myfunc">',
            rendered)

    def test_ref_break_for_python(self):
        """Testing ref_utils reference line break for Python references"""
        rendered = self.render_doc(
            '.. py:function:: abc.def.myfunc\n'
            '\n'
            '   Description.\n'
            '\n'
            'Link to :py:class:`abc.def\n'
            '.myfunc`.'
        )

        self.assertIn(
            '<a class="reference internal" href="#abc.def.myfunc"'
            ' title="abc.def.myfunc">',
            rendered)



"""Configures pytest for beanbag-docutils.

Version Added:
    2.1
"""

import django
import sphinx


def pytest_report_header(config):
    """Return information for the report header.

    This will log the version of Sphinx.

    Args:
        config (object):
            The pytest configuration object.

    Returns:
        list of str:
        The report header entries to log.
    """
    return [
        'Sphinx: %s' % sphinx.__version__,
        'Django: %s' % django.get_version(),
    ]



"""Sphinx extension for Retina images.

.. deprecated:: 2.1

   We recommend using :py:mod:`beanbag_docutils.sphinx.ext.image_srcsets`
   instead. This extension will be removed in a future release.

This extension goes through all the images Sphinx will provide in _images and
checks if Retina versions are available. If there are any, they will be copied
as well.


Setup
=====

To use this, you just need to add the extension in :file:`conf.py`::

    extensions = [
        ...
        'beanbag_docutils.sphinx.ext.retina_images',
        ...
    ]


Configuration
=============

``retina_suffixes``:
    A list of suffix identifiers for Retina images. Each of these go after
    the filename and before the extension. This defaults to ``['@2x', '@3x']``.
"""

import os


def add_high_dpi_images(app, env):
    """Add high-DPI images to the list of bundled images.

    Any image that has a "@2x" version will be included in the output
    directory for the docs.

    Args:
        app (sphinx.application.Sphinx):
            The Sphinx application to register roles and configuration on.

        env (sphinx.environment.BuildEnvironment):
            The build environment for the generated docs.
    """
    suffixes = app.config['retina_suffixes']
    retina_images = []

    for full_path, (docnames, filename) in env.images.items():
        base, ext = os.path.splitext(full_path)

        for suffix in suffixes:
            src_retina_path = '%s%s%s' % (base, suffix, ext)

            if os.path.exists(src_retina_path):
                base, ext = os.path.splitext(filename)
                dest_retina_name = '%s%s%s' % (base, suffix, ext)

                retina_images += [
                    (docname, src_retina_path, dest_retina_name)
                    for docname in docnames
                ]

    for docname, src_path, dest_name in retina_images:
        # Emulate add_file(), but give greater control over the filenames.
        # Ideally we wouldn't dive into internals of _existing, but we should
        # be able to adjust this easily enough for any changes that may be
        # made.
        env.images[src_path] = (set([docname]), dest_name)
        env.images._existing.add(dest_name)


def collect_pages(app):
    """Collect high-DPI images for use in HTML pages.

    This will go through the images referenced in a document for an HTML page
    and add any high-DPI versions previously found in
    :py:func:`add_high_dpi_images` to the list of images to collect for the
    page.

    Args:
        app (sphinx.application.Sphinx):
            The Sphinx application to register roles and configuration on.

    Returns:
        list:
        An empty list (indicating no additional HTML pages are collected).
    """
    suffixes = app.config['retina_suffixes']
    new_images = {}

    for full_path, basename in app.builder.images.items():
        base, ext = os.path.splitext(full_path)

        for suffix in suffixes:
            retina_path = '%s%s%s' % (base, suffix, ext)

            if retina_path in app.env.images:
                new_images[retina_path] = app.env.images[retina_path][1]

    app.builder.images.update(new_images)

    return []


def setup(app):
    """Set up the Sphinx extension.

    This listens for the events needed to collect and bundle high-DPI
    images.

    Args:
        app (sphinx.application.Sphinx):
            The Sphinx application to listen to events on.
    """
    app.add_config_value('retina_suffixes', ['@2x', '@3x'], True)

    app.connect('env-updated', add_high_dpi_images)
    app.connect('html-collect-pages', collect_pages)



"""Sphinx extension to add a :rst:role:`http` role for docs.

This extension makes it easy to reference HTTP status codes.


Setup
=====

To use this, you just need to add the extension in :file:`conf.py`::

    extensions = [
        ...
        'beanbag_docutils.sphinx.ext.http_role',
        ...
    ]


Roles
=====

.. rst:directive:: http-status-codes-format

   Specifies a new format to use by default for any :rst:role:`http` roles.
   This should include ``%(code)s`` for the numeric code and ``%(name)s``
   for the name of the HTTP status code.

   Call this again without an argument to use the default format.


.. rst:role:: http

   References an HTTP status code, expanding to the full status name and
   linking to documentation on the status in the process.


Configuration
=============

``http_status_codes_format``
    The format string used for the titles for HTTP status codes. This
    defaults to ``HTTP %(code)s %(format)s`` and can be temporarily
    overridden using :rst:dir:`http-status-codes-format`.

``http_status_codes_url``
    The location of the docs for the status codes. This expects a string with a
    ``%s``, which will be replaced by the numeric HTTP status code.
"""

from docutils import nodes
from docutils.parsers.rst import Directive

from sphinx.util.nodes import split_explicit_title


DEFAULT_HTTP_STATUS_CODES_URL = \
    'https://wikipedia.org/wiki/List_of_HTTP_status_codes#%s'

DEFAULT_HTTP_STATUS_CODES_FORMAT = 'HTTP %(code)s %(name)s'

HTTP_STATUS_CODES = {
    100: 'Continue',
    101: 'Switching Protocols',
    102: 'Processing',
    200: 'OK',
    201: 'Created',
    202: 'Accepted',
    203: 'Non-Authoritative Information',
    204: 'No Content',
    205: 'Reset Content',
    206: 'Partial Content',
    207: 'Multi-Status',
    208: 'Already Reported',
    300: 'Multiple Choices',
    301: 'Moved Permanently',
    302: 'Found',
    303: 'See Other',
    304: 'Not Modified',
    305: 'Use Proxy',
    306: 'Switch Proxy',
    307: 'Temporary Redirect',
    308: 'Permanent Redirect',
    400: 'Bad Request',
    401: 'Unauthorized',
    402: 'Payment Required',
    403: 'Forbidden',
    404: 'Not Found',
    405: 'Method Not Allowed',
    406: 'Not Acceptable',
    407: 'Proxy Authentication Required',
    408: 'Request Timeout',
    409: 'Conflict',
    410: 'Gone',
    411: 'Length Required',
    412: 'Precondition Failed',
    413: 'Request Entity Too Large',
    414: 'Request-URI Too Long',
    415: 'Unsupported Media Type',
    416: 'Requested Range Not Satisfiable',
    417: 'Expectation Failed',
    418: "I'm a teapot",
    422: 'Unprocessable Entity',
    423: 'Locked',
    424: 'Failed Dependency',
    425: 'Unordered Collection',
    426: 'Upgrade Required',
    428: 'Precondition Required',
    429: 'Too Many Requests',
    431: 'Request Header Fields Too Large',
    444: 'No Response',
    449: 'Retry With',
    450: 'Blocked by Windows Parental Controls',
    451: 'Unavailable For Legal Reasons',
    499: 'Client Closed Request',
    500: 'Internal Server Error',
    501: 'Not Implemented',
    502: 'Bad Gateway',
    503: 'Service Unavailable',
    504: 'Gateway Timeout',
    505: 'HTTP Version Not Supported',
    506: 'Variant Also Negotiates',
    507: 'Insufficient Storage',
    508: 'Loop Detected',
    509: 'Bandwidth Limit Exceeded',
    510: 'Not Extended',
    511: 'Network Authentication Required',
    598: 'Network Read Timeout Error',
    599: 'Network Connect Timeout Error',
}


class SetStatusCodesFormatDirective(Directive):
    """Specifies the format to use for the ``:http:`` role's text."""

    required_arguments = 0
    optional_arguments = 1
    final_argument_whitespace = True

    def run(self):
        """Run the directive.

        Returns:
            list:
            An empty list, always.
        """
        temp_data = self.state.document.settings.env.temp_data

        if len(self.arguments):
            temp_data['http-status-codes-format'] = self.arguments[0]
        else:
            temp_data.pop('http-status-codes-format', None)

        return []


def http_role(role, rawtext, text, linenum, inliner, options={}, content=[]):
    """Implementation of the :rst:role:`http` role.

    This is responsible for converting a HTTP status code to link pointing to
    the status documentation, with the full text for the status name.

    Args:
        rawtext (unicode):
            The raw text for the entire role.

        text (unicode):
            The interpreted text content.

        linenum (int):
            The current line number.

        inliner (docutils.parsers.rst.states.Inliner):
            The inliner used for error reporting and document tree access.

        options (dict):
            Options passed for the role. This is unused.

        content (list of unicode):
            The list of strings containing content for the role directive.
            This is unused.

    Returns:
        tuple:
        The result of the role. It's a tuple containing two items:

        1) A single-item list with the resulting node.
        2) A single-item list with the error message (if any).
    """
    has_explicit_title, title, target = split_explicit_title(text)

    try:
        status_code = int(target)

        if status_code not in HTTP_STATUS_CODES:
            raise ValueError
    except ValueError:
        msg = inliner.reporter.error(
            'HTTP status code must be a valid HTTP status; '
            '"%s" is invalid.' % target,
            line=linenum)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]

    env = inliner.document.settings.env
    http_status_codes_url = env.config.http_status_codes_url

    if not http_status_codes_url or '%s' not in http_status_codes_url:
        msg = inliner.reporter.error('http_status_codes_url must be '
                                     'configured.',
                                     line=linenum)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]

    ref = http_status_codes_url % status_code

    if has_explicit_title:
        status_code_text = title
    else:
        http_status_codes_format = (
            env.temp_data.get('http-status-codes-format') or
            env.config.http_status_codes_format
        )
        status_code_text = http_status_codes_format % {
            'code': status_code,
            'name': HTTP_STATUS_CODES[status_code],
        }

    node = nodes.reference(rawtext, status_code_text, refuri=ref, **options)

    return [node], []


def setup(app):
    """Set up the Sphinx extension.

    This registers the :rst:role:`http` role,
    :rst:dir:`http-status-codes-format` directive, and the configuration
    settings for specifying the format and URL for linking to HTTP status
    codes.

    Args:
        app (sphinx.application.Sphinx):
            The Sphinx application to register roles and configuration on.
    """
    app.add_config_value('http_status_codes_format',
                         DEFAULT_HTTP_STATUS_CODES_FORMAT,
                         True)
    app.add_config_value('http_status_codes_url',
                         DEFAULT_HTTP_STATUS_CODES_URL,
                         True)

    app.add_directive('http-status-codes-format',
                      SetStatusCodesFormatDirective)
    app.add_role('http', http_role)



"""Unit tests for beanbag_docutils.sphinx.ext.extlinks."""

from sphinx import version_info as sphinx_version_info

from beanbag_docutils.sphinx.ext.tests.testcase import SphinxExtTestCase


if sphinx_version_info[:2] >= (4, 0):
    _bug_caption = 'bug%s'
else:
    _bug_caption = 'bug'


class ExtLinksTests(SphinxExtTestCase):
    """Unit tests for extlinks."""

    config = {
        'extlinks': {
            'bug': ('https://bugs.example.com/%s?#main', _bug_caption),
        },
    }

    extensions = [
        'beanbag_docutils.sphinx.ext.extlinks',
    ]

    def test_extlink_with_anchor(self):
        """Testing :<extlink>:`...` role with mapped URL containing anchor"""
        self.assertEqual(
            self.render_doc(':bug:`123`'),
            '<p><a class="reference external"'
            ' href="https://bugs.example.com/123?#main">bug123</a></p>')



"""Unit tests for beanbag_docutils.sphinx.ext.github"""

import kgb

from beanbag_docutils.sphinx.ext.github import (clear_github_linkcode_caches,
                                                github_linkcode_resolve,
                                                _run_git)
from beanbag_docutils.sphinx.ext.tests.testcase import SphinxExtTestCase
from beanbag_docutils.sphinx.ext.tests.testdata import github_linkcode_module


class GitHubLinkCodeResolveTests(kgb.SpyAgency, SphinxExtTestCase):
    """Unit tests for github_linkcode_resolve"""

    SRC_MODULE = github_linkcode_module.__name__

    def tearDown(self):
        super(GitHubLinkCodeResolveTests, self).tearDown()

        clear_github_linkcode_caches()

    def test_with_link_to_class(self):
        """Testing github_linkcode_resolve with linking to class"""
        self.spy_on(_run_git, op=kgb.SpyOpMatchInOrder([
            {
                'args': (['fetch', 'origin', 'mybranch', 'origin'],),
                'call_original': False,
            },
            {
                'args': (['branch', '-rv', '--contains', 'mybranch'],),
                'op': kgb.SpyOpReturn(
                    b'  origin/mybranch abcd123 Here is a commit.\n'
                ),
            },
            {
                'args': (['log', '--pretty=format:%H', '...abcd123'],),
                'op': kgb.SpyOpReturn(
                    b'157ac365d792c79987966e0152d16d6d1526b24d\n'
                ),
            },
            {
                'args': (['rev-parse', 'mybranch'],),
                'op': kgb.SpyOpReturn(
                    b'55ca6286e3e4f4fba5d0448333fa99fc5a404a73\n'
                ),
            },
        ]))

        url = github_linkcode_resolve(
            domain='py',
            info={
                'module': self.SRC_MODULE,
                'fullname': 'ClassB',
            },
            github_org_id='beanbaginc',
            github_repo_id='beanbag_docutils',
            branch='mybranch')

        self.assertIsNotNone(url)
        self.assertIsInstance(url, str)
        self.assertEqual(
            url,
            'https://github.com/beanbaginc/beanbag_docutils/blob/'
            '55ca6286e3e4f4fba5d0448333fa99fc5a404a73/beanbag_docutils/'
            'sphinx/ext/tests/testdata/github_linkcode_module.py#L9')

        self.assertSpyCallCount(_run_git, 4)

    def test_with_link_to_function(self):
        """Testing github_linkcode_resolve with linking to function"""
        self.spy_on(_run_git, op=kgb.SpyOpMatchInOrder([
            {
                'args': (['fetch', 'origin', 'mybranch', 'origin'],),
                'call_original': False,
            },
            {
                'args': (['branch', '-rv', '--contains', 'mybranch'],),
                'op': kgb.SpyOpReturn(
                    b'  origin/mybranch abcd123 Here is a commit.\n'
                ),
            },
            {
                'args': (['log', '--pretty=format:%H', '...abcd123'],),
                'op': kgb.SpyOpReturn(
                    b'157ac365d792c79987966e0152d16d6d1526b24d\n'
                ),
            },
            {
                'args': (['rev-parse', 'mybranch'],),
                'op': kgb.SpyOpReturn(
                    b'55ca6286e3e4f4fba5d0448333fa99fc5a404a73\n'
                ),
            },
        ]))

        url = github_linkcode_resolve(
            domain='py',
            info={
                'module': self.SRC_MODULE,
                'fullname': 'my_func',
            },
            github_org_id='beanbaginc',
            github_repo_id='beanbag_docutils',
            branch='mybranch')

        self.assertIsNotNone(url)
        self.assertIsInstance(url, str)
        self.assertEqual(
            url,
            'https://github.com/beanbaginc/beanbag_docutils/blob/'
            '55ca6286e3e4f4fba5d0448333fa99fc5a404a73/beanbag_docutils/'
            'sphinx/ext/tests/testdata/github_linkcode_module.py#L21')

        self.assertSpyCallCount(_run_git, 4)

    def test_with_link_to_class_attr(self):
        """Testing github_linkcode_resolve with linking to class attribute"""
        self.spy_on(_run_git, op=kgb.SpyOpMatchInOrder([
            {
                'args': (['fetch', 'origin', 'mybranch', 'origin'],),
                'call_original': False,
            },
            {
                'args': (['branch', '-rv', '--contains', 'mybranch'],),
                'op': kgb.SpyOpReturn(
                    b'  origin/mybranch abcd123 Here is a commit.\n'
                ),
            },
            {
                'args': (['log', '--pretty=format:%H', '...abcd123'],),
                'op': kgb.SpyOpReturn(
                    b'157ac365d792c79987966e0152d16d6d1526b24d\n'
                ),
            },
            {
                'args': (['rev-parse', 'mybranch'],),
                'op': kgb.SpyOpReturn(
                    b'55ca6286e3e4f4fba5d0448333fa99fc5a404a73\n'
                ),
            },
        ]))

        url = github_linkcode_resolve(
            domain='py',
            info={
                'module': self.SRC_MODULE,
                'fullname': 'ClassB.my_var',
            },
            github_org_id='beanbaginc',
            github_repo_id='beanbag_docutils',
            branch='mybranch')

        self.assertIsNotNone(url)
        self.assertIsInstance(url, str)
        self.assertEqual(
            url,
            'https://github.com/beanbaginc/beanbag_docutils/blob/'
            '55ca6286e3e4f4fba5d0448333fa99fc5a404a73/beanbag_docutils/'
            'sphinx/ext/tests/testdata/github_linkcode_module.py#L12')

        self.assertSpyCallCount(_run_git, 4)

    def test_with_link_to_variable(self):
        """Testing github_linkcode_resolve with linking to global variable"""
        self.spy_on(_run_git, op=kgb.SpyOpMatchInOrder([
            {
                'args': (['fetch', 'origin', 'mybranch', 'origin'],),
                'call_original': False,
            },
            {
                'args': (['branch', '-rv', '--contains', 'mybranch'],),
                'op': kgb.SpyOpReturn(
                    b'  origin/mybranch abcd123 Here is a commit.\n'
                ),
            },
            {
                'args': (['log', '--pretty=format:%H', '...abcd123'],),
                'op': kgb.SpyOpReturn(
                    b'157ac365d792c79987966e0152d16d6d1526b24d\n'
                ),
            },
            {
                'args': (['rev-parse', 'mybranch'],),
                'op': kgb.SpyOpReturn(
                    b'55ca6286e3e4f4fba5d0448333fa99fc5a404a73\n'
                ),
            },
        ]))

        url = github_linkcode_resolve(
            domain='py',
            info={
                'module': self.SRC_MODULE,
                'fullname': 'my_var',
            },
            github_org_id='beanbaginc',
            github_repo_id='beanbag_docutils',
            branch='mybranch')

        self.assertIsNotNone(url)
        self.assertIsInstance(url, str)
        self.assertEqual(
            url,
            'https://github.com/beanbaginc/beanbag_docutils/blob/'
            '55ca6286e3e4f4fba5d0448333fa99fc5a404a73/beanbag_docutils/'
            'sphinx/ext/tests/testdata/github_linkcode_module.py#L25')

        self.assertSpyCallCount(_run_git, 4)

    def test_with_link_to_method(self):
        """Testing github_linkcode_resolve with linking to method"""
        self.spy_on(_run_git, op=kgb.SpyOpMatchInOrder([
            {
                'args': (['fetch', 'origin', 'mybranch', 'origin'],),
                'call_original': False,
            },
            {
                'args': (['branch', '-rv', '--contains', 'mybranch'],),
                'op': kgb.SpyOpReturn(
                    b'  origin/mybranch abcd123 Here is a commit.\n'
                ),
            },
            {
                'args': (['log', '--pretty=format:%H', '...abcd123'],),
                'op': kgb.SpyOpReturn(
                    b'157ac365d792c79987966e0152d16d6d1526b24d\n'
                ),
            },
            {
                'args': (['rev-parse', 'mybranch'],),
                'op': kgb.SpyOpReturn(
                    b'55ca6286e3e4f4fba5d0448333fa99fc5a404a73\n'
                ),
            },
        ]))

        url = github_linkcode_resolve(
            domain='py',
            info={
                'module': self.SRC_MODULE,
                'fullname': 'ClassB.do_thing',
            },
            github_org_id='beanbaginc',
            github_repo_id='beanbag_docutils',
            branch='mybranch')

        self.assertIsNotNone(url)
        self.assertIsInstance(url, str)
        self.assertEqual(
            url,
            'https://github.com/beanbaginc/beanbag_docutils/blob/'
            '55ca6286e3e4f4fba5d0448333fa99fc5a404a73/beanbag_docutils/'
            'sphinx/ext/tests/testdata/github_linkcode_module.py#L17')

        self.assertSpyCallCount(_run_git, 4)

    def test_with_link_to_object_attr_not_in_code(self):
        """Testing github_linkcode_resolve with linking to an attribute on an
        object not defined in code
        """
        self.spy_on(_run_git, op=kgb.SpyOpMatchInOrder([
            {
                'args': (['fetch', 'origin', 'mybranch', 'origin'],),
                'call_original': False,
            },
            {
                'args': (['branch', '-rv', '--contains', 'mybranch'],),
                'op': kgb.SpyOpReturn(
                    b'  origin/mybranch abcd123 Here is a commit.\n'
                ),
            },
            {
                'args': (['log', '--pretty=format:%H', '...abcd123'],),
                'op': kgb.SpyOpReturn(
                    b'157ac365d792c79987966e0152d16d6d1526b24d\n'
                ),
            },
            {
                'args': (['rev-parse', 'mybranch'],),
                'op': kgb.SpyOpReturn(
                    b'55ca6286e3e4f4fba5d0448333fa99fc5a404a73\n'
                ),
            },
        ]))

        url = github_linkcode_resolve(
            domain='py',
            info={
                'module': self.SRC_MODULE,
                'fullname': 'MyNamedTuple.__match_args__',
            },
            github_org_id='beanbaginc',
            github_repo_id='beanbag_docutils',
            branch='mybranch')

        self.assertIsNotNone(url)
        self.assertIsInstance(url, str)
        self.assertEqual(
            url,
            'https://github.com/beanbaginc/beanbag_docutils/blob/'
            '55ca6286e3e4f4fba5d0448333fa99fc5a404a73/beanbag_docutils/'
            'sphinx/ext/tests/testdata/github_linkcode_module.py#L27')

        self.assertSpyCallCount(_run_git, 4)

    def test_with_allowed_modules_and_found(self):
        """Testing github_linkcode_resolve with allowed_modules and module
        found
        """
        self.spy_on(_run_git, op=kgb.SpyOpMatchInOrder([
            {
                'args': (['fetch', 'origin', 'mybranch', 'origin'],),
                'call_original': False,
            },
            {
                'args': (['branch', '-rv', '--contains', 'mybranch'],),
                'op': kgb.SpyOpReturn(
                    b'  origin/mybranch abcd123 Here is a commit.\n'
                ),
            },
            {
                'args': (['log', '--pretty=format:%H', '...abcd123'],),
                'op': kgb.SpyOpReturn(
                    b'157ac365d792c79987966e0152d16d6d1526b24d\n'
                ),
            },
            {
                'args': (['rev-parse', 'mybranch'],),
                'op': kgb.SpyOpReturn(
                    b'55ca6286e3e4f4fba5d0448333fa99fc5a404a73\n'
                ),
            },
        ]))

        url = github_linkcode_resolve(
            domain='py',
            info={
                'module': self.SRC_MODULE,
                'fullname': 'ClassB.do_thing',
            },
            github_org_id='beanbaginc',
            github_repo_id='beanbag_docutils',
            branch='mybranch',
            allowed_module_names=['beanbag_docutils'])

        self.assertIsNotNone(url)
        self.assertIsInstance(url, str)
        self.assertEqual(
            url,
            'https://github.com/beanbaginc/beanbag_docutils/blob/'
            '55ca6286e3e4f4fba5d0448333fa99fc5a404a73/beanbag_docutils/'
            'sphinx/ext/tests/testdata/github_linkcode_module.py#L17')

        self.assertSpyCallCount(_run_git, 4)

    def test_with_allowed_modules_and_not_found(self):
        """Testing github_linkcode_resolve with allowed_modules and module
        not found
        """
        self.spy_on(_run_git, call_original=False)

        url = github_linkcode_resolve(
            domain='py',
            info={
                'module': self.SRC_MODULE,
                'fullname': 'ClassB.do_thing',
            },
            github_org_id='beanbaginc',
            github_repo_id='beanbag_docutils',
            branch='mybranch',
            allowed_module_names=['something_else'])

        self.assertIsNone(url)
        self.assertSpyCallCount(_run_git, 0)

    def test_with_source_prefix(self):
        """Testing github_linkcode_resolve with source_prefix"""
        self.spy_on(_run_git, op=kgb.SpyOpMatchInOrder([
            {
                'args': (['fetch', 'origin', 'mybranch', 'origin'],),
                'call_original': False,
            },
            {
                'args': (['branch', '-rv', '--contains', 'mybranch'],),
                'op': kgb.SpyOpReturn(
                    b'  origin/mybranch abcd123 Here is a commit.\n'
                ),
            },
            {
                'args': (['log', '--pretty=format:%H', '...abcd123'],),
                'op': kgb.SpyOpReturn(
                    b'157ac365d792c79987966e0152d16d6d1526b24d\n'
                ),
            },
            {
                'args': (['rev-parse', 'mybranch'],),
                'op': kgb.SpyOpReturn(
                    b'55ca6286e3e4f4fba5d0448333fa99fc5a404a73\n'
                ),
            },
        ]))

        url = github_linkcode_resolve(
            domain='py',
            info={
                'module': self.SRC_MODULE,
                'fullname': 'ClassB.do_thing',
            },
            github_org_id='beanbaginc',
            github_repo_id='beanbag_docutils',
            branch='mybranch',
            source_prefix='foo/bar/')

        self.assertIsNotNone(url)
        self.assertIsInstance(url, str)
        self.assertEqual(
            url,
            'https://github.com/beanbaginc/beanbag_docutils/blob/'
            '55ca6286e3e4f4fba5d0448333fa99fc5a404a73/foo/bar/'
            'beanbag_docutils/sphinx/ext/tests/testdata/'
            'github_linkcode_module.py#L17')

        self.assertSpyCallCount(_run_git, 4)

    def test_with_github_url(self):
        """Testing github_linkcode_resolve with custom github_url="""
        self.spy_on(_run_git, op=kgb.SpyOpMatchInOrder([
            {
                'args': (['fetch', 'origin', 'mybranch', 'origin'],),
                'call_original': False,
            },
            {
                'args': (['branch', '-rv', '--contains', 'mybranch'],),
                'op': kgb.SpyOpReturn(
                    b'  origin/mybranch abcd123 Here is a commit.\n'
                ),
            },
            {
                'args': (['log', '--pretty=format:%H', '...abcd123'],),
                'op': kgb.SpyOpReturn(
                    b'157ac365d792c79987966e0152d16d6d1526b24d\n'
                ),
            },
            {
                'args': (['rev-parse', 'mybranch'],),
                'op': kgb.SpyOpReturn(
                    b'55ca6286e3e4f4fba5d0448333fa99fc5a404a73\n'
                ),
            },
        ]))

        url = github_linkcode_resolve(
            domain='py',
            info={
                'module': self.SRC_MODULE,
                'fullname': 'ClassB',
            },
            github_org_id='beanbaginc',
            github_repo_id='beanbag_docutils',
            branch='mybranch',
            github_url='https://dev.example.com/github')

        self.assertIsNotNone(url)
        self.assertIsInstance(url, str)
        self.assertEqual(
            url,
            'https://dev.example.com/github/beanbaginc/beanbag_docutils/blob/'
            '55ca6286e3e4f4fba5d0448333fa99fc5a404a73/beanbag_docutils/'
            'sphinx/ext/tests/testdata/github_linkcode_module.py#L9')

        self.assertSpyCallCount(_run_git, 4)

    def test_with_github_url_trailing_slash(self):
        """Testing github_linkcode_resolve with custom github_url= with
        trailing slash
        """
        self.spy_on(_run_git, op=kgb.SpyOpMatchInOrder([
            {
                'args': (['fetch', 'origin', 'mybranch', 'origin'],),
                'call_original': False,
            },
            {
                'args': (['branch', '-rv', '--contains', 'mybranch'],),
                'op': kgb.SpyOpReturn(
                    b'  origin/mybranch abcd123 Here is a commit.\n'
                ),
            },
            {
                'args': (['log', '--pretty=format:%H', '...abcd123'],),
                'op': kgb.SpyOpReturn(
                    b'157ac365d792c79987966e0152d16d6d1526b24d\n'
                ),
            },
            {
                'args': (['rev-parse', 'mybranch'],),
                'op': kgb.SpyOpReturn(
                    b'55ca6286e3e4f4fba5d0448333fa99fc5a404a73\n'
                ),
            },
        ]))

        url = github_linkcode_resolve(
            domain='py',
            info={
                'module': self.SRC_MODULE,
                'fullname': 'ClassB',
            },
            github_org_id='beanbaginc',
            github_repo_id='beanbag_docutils',
            branch='mybranch',
            github_url='https://dev.example.com/github/')

        self.assertIsNotNone(url)
        self.assertIsInstance(url, str)
        self.assertEqual(
            url,
            'https://dev.example.com/github/beanbaginc/beanbag_docutils/blob/'
            '55ca6286e3e4f4fba5d0448333fa99fc5a404a73/beanbag_docutils/'
            'sphinx/ext/tests/testdata/github_linkcode_module.py#L9')

        self.assertSpyCallCount(_run_git, 4)

    def test_with_multiple_refs(self):
        """Testing github_linkcode_resolve with multiple refs for branch"""
        self.spy_on(_run_git, op=kgb.SpyOpMatchInOrder([
            {
                'args': (['fetch', 'origin', 'mybranch', 'origin'],),
                'call_original': False,
            },
            {
                'args': (['branch', '-rv', '--contains', 'mybranch'],),
                'op': kgb.SpyOpReturn(
                    b'  origin/mybranch abcd123 Here is a commit.\n'
                    b'  origin/mybranch-dev def4567 Branched for dev.\n'
                ),
            },
            {
                'args': (['log', '--pretty=format:%H', '...abcd123'],),
                'op': kgb.SpyOpReturn(
                    b'abcd1235d792c79987966e0152d16d6d1526b24d\n'
                    b'55ca6286e3e4f4fba5d0448333fa99fc5a404a73\n'
                ),
            },
            {
                'args': (['log', '--pretty=format:%H', '...def4567'],),
                'op': kgb.SpyOpReturn(
                    b'def45675d792c79987966e0152d16d6d1526b24d\n'
                ),
            },
            {
                'args': (['rev-parse', 'mybranch-dev'],),
                'op': kgb.SpyOpReturn(
                    b'55ca6286e3e4f4fba5d0448333fa99fc5a404a73\n'
                ),
            },
        ]))

        url = github_linkcode_resolve(
            domain='py',
            info={
                'module': self.SRC_MODULE,
                'fullname': 'ClassB.do_thing',
            },
            github_org_id='beanbaginc',
            github_repo_id='beanbag_docutils',
            branch='mybranch',
            source_prefix='foo/bar/')

        self.assertIsNotNone(url)
        self.assertIsInstance(url, str)
        self.assertEqual(
            url,
            'https://github.com/beanbaginc/beanbag_docutils/blob/'
            '55ca6286e3e4f4fba5d0448333fa99fc5a404a73/foo/bar/'
            'beanbag_docutils/sphinx/ext/tests/testdata/'
            'github_linkcode_module.py#L17')

        self.assertSpyCallCount(_run_git, 5)

    def test_with_tracking_branch_not_found(self):
        """Testing github_linkcode_resolve with tracking branch not found"""
        self.spy_on(_run_git, op=kgb.SpyOpMatchInOrder([
            {
                'args': (['fetch', 'origin', 'mybranch', 'origin'],),
                'call_original': False,
            },
            {
                'args': (['branch', '-rv', '--contains', 'mybranch'],),
                'op': kgb.SpyOpReturn(b''),
            },
        ]))

        url = github_linkcode_resolve(
            domain='py',
            info={
                'module': self.SRC_MODULE,
                'fullname': 'ClassB.do_thing',
            },
            github_org_id='beanbaginc',
            github_repo_id='beanbag_docutils',
            branch='mybranch',
            allowed_module_names=['beanbag_docutils'])

        self.assertIsNone(url)
        self.assertSpyCallCount(_run_git, 2)

    def test_with_non_py_domain(self):
        """Testing github_linkcode_resolve with non-"py" domain"""
        self.spy_on(_run_git, call_original=False)

        url = github_linkcode_resolve(
            domain='js',
            info={
                'module': self.SRC_MODULE,
                'fullname': 'ClassB.do_thing',
            },
            github_org_id='beanbaginc',
            github_repo_id='beanbag_docutils',
            branch='mybranch',
            source_prefix='foo/bar/')

        self.assertIsNone(url)

    def test_with_no_module_name(self):
        """Testing github_linkcode_resolve with no module name"""
        self.spy_on(_run_git, call_original=False)

        url = github_linkcode_resolve(
            domain='py',
            info={
                'module': '',
                'fullname': 'ClassB.do_thing',
            },
            github_org_id='beanbaginc',
            github_repo_id='beanbag_docutils',
            branch='mybranch',
            source_prefix='foo/bar/')

        self.assertIsNone(url)

    def test_with_module_not_found(self):
        """Testing github_linkcode_resolve with module not found"""
        self.spy_on(_run_git, call_original=False)

        url = github_linkcode_resolve(
            domain='py',
            info={
                'module': 'beanbag_docutils.xxx',
                'fullname': 'ClassB.do_thing',
            },
            github_org_id='beanbaginc',
            github_repo_id='beanbag_docutils',
            branch='mybranch',
            source_prefix='foo/bar/')

        self.assertIsNone(url)

    def test_with_fullname_not_found(self):
        """Testing github_linkcode_resolve with fullname object not found"""
        self.spy_on(_run_git, call_original=False)

        url = github_linkcode_resolve(
            domain='py',
            info={
                'module': self.SRC_MODULE,
                'fullname': 'ClassB.invalid_function',
            },
            github_org_id='beanbaginc',
            github_repo_id='beanbag_docutils',
            branch='mybranch',
            source_prefix='foo/bar/')

        self.assertIsNone(url)



# coding: utf-8
"""Unit tests for beanbag_docutils.sphinx.ext.metadata.

Version Added:
    2.2
"""

import json
from unittest import SkipTest

from docutils import nodes, __version__ as DOCUTILS_VERSION

from beanbag_docutils.sphinx.ext import metadata
from beanbag_docutils.sphinx.ext.tests.testcase import SphinxExtTestCase


class MetadataTests(SphinxExtTestCase):
    """Unit tests for the metadata extension."""

    extensions = [
        metadata.__name__,
    ]

    @classmethod
    def setUpClass(cls):
        super(MetadataTests, cls).setUpClass()

        if not hasattr(nodes, 'meta'):
            raise SkipTest('Not supported on docutils %s' % DOCUTILS_VERSION)

    def test_with_single_value(self):
        """Testing metadata extension with single value for key"""
        data = json.loads(self.render_doc(
            '.. meta::\n'
            '   :my_key: my value!\n',
            builder_name='json'
        ))

        self.assertEqual(
            data.get('meta'),
            {
                'my_key': {
                    'content': 'my value!',
                },
            })

    def test_with_multiple_values(self):
        """Testing metadata extension with multiple values for key"""
        data = json.loads(self.render_doc(
            '.. meta::\n'
            '   :description lang=en: hi!\n'
            '   :description lang=ja: こんにちは\n',
            builder_name='json'
        ))

        self.assertEqual(
            data.get('meta'),
            {
                'description': [
                    {
                        'content': 'hi!',
                        'lang': 'en',
                    },
                    {
                        'content': 'こんにちは',
                        'lang': 'ja',
                    },
                ],
            })
